# Resource object code (Python 3)
# Created by: object code
# Created by: The Resource Compiler for Qt version 5.15.0
# WARNING! All changes made in this file will be lost!

from PySide2 import QtCore

qt_resource_data = b"\
\x00\x00\x07*\
\x00\
\x00& x\x9c\xddZ\xddo\xe3\xb8\x11\x7f\xcf_\xa1\
j_vQ\x8b\xe2\xf0\x9b\xae\x9d\x03z\x8b\x03\xee\xb5\
\xbd\xc3=\x16\x8a\xc4\xd8\xea\xca\x92!\xc9qr\x7f}\
\x87\xb2%\x7f(.\x1c \x05\x5c+Yl\xc8\x19R\
3?\xce'\xed\xd9O\xaf\xab\x22xqu\x93W\xe5\
<\x04B\xc3\xc0\x95i\x95\xe5\xe5b\x1e\xfe\xfe\xdb/\
\x91\x09\x83\xa6M\xca,)\xaa\xd2\xcd\xc3\xb2\x0a\x7fz\
|\x98\xfd%\x8a\x82\x9fk\x97\xb4.\x0b\xb6y\xbb\x0c\
~-\x7f4i\xb2v\xc1\xd7e\xdb\xae\xa7q\xbc\xdd\
nI\xbe\x9f$U\xbd\x88\xbf\x05Q\xf4\xf8\xf00k\
^\x16\x0fA\x10\xe0{\xcbf\x9a\xa5\xf3p\xbf`\xbd\
\xa9\x8b\x8e1KcW\xb8\x95+\xdb&\x06\x02qx\
`O\x0f\xec\xa9\x7f{\xfe\xe2\xd2j\xb5\xaa\xca\xa6[\
Y6_\x8e\x98\xeb\xecy\xe0\xf6\xd2ly\xc7\x04\xd6\
\xda\x98\xb2\x98\xb1\x089\xa2\xe6\xadl\x93\xd7\xe8t)\
\xca\xf8\xdeRF)\x8d\x91v\xe0\xbc\x8ek\xda \xa0\
k\xfc7\xb0\xf7\x13\xa4\xa96u\xea\x9eq\x9d#\xa5\
k\xe3\xef\xbf}\x1f\x88\x11%Y\x9b\x1dm\xd3\xe3y\
\xf2\xd6\x13\x90\xcbd\xe5\x9au\x92\xba&\xee\xe7\xbb\xf5\
\xdb<k\x97\xf3P\x98n\xb4t\xf9b\xd9\x0e\xc3\x97\
\xdcm\xff^\xbd\xceC\x1a\xd0@\x18\xfc%\xd4?\xd0\
Q\xf3l\x1e\xa26l\xc7z\xb0\x94=u\xff\x96\xe9\
@\xa1\xc42\xc2\x82\xaf2\xe5\xce\xd0l\x120\x0a:\
\xa2&\xa2\xea[\xb7\xa4Wo\x9aU\xa9\x97w\x1e&\
\x9b\xb6Z\xe1i\xa6\xff*\x92\xb7j\xd3\x12\x8f\xde#\
\xf2\xce2\xf7\xdc\xf85;1\xfcH\x84A\xdc\x91\x86\
m\xfc\x1e\x99\xd7\xe1\xc0\xf8\x944;\xbd\x83`\x9d,\
\xd0F\x8a\xaa\x9e\x87_\x9e\xbbgOx\xaa\xea\xcc\xd5\
=Iu\xcf\x09\xa9B\x1c\xf3\xf6m\xe7\x15\xfb\xbd{\
m\xfd\xae\x03\x9d\xbeOo\x96IVm\xe7!;'\
\xfeYU+\x04\x9fh\xa3\x844\xfc\x9c\x9c\xe2ID\
x\x06\x96\x82\xb2fD\xc5\x172 \x9aq&FK\
\x11\xd1\x8dw\x9chS\xe6-\x1a\xe7\xfau\xb4|S\
\xd7\x9e\x01\x81v\xa8\xf7\x82\x0b\x01{\x9efYm\x17\
\xb5\x87\xef9)\x06\xfc.\xee\xb4\xcdKT/\xda[\
\x16\x185\x92f\xcf\xd1[\x1bP&/\xb0\xa0\xc6R\
_\xa0y}/\xd1V\xc9k\xbe\xca\xfft(3\xf4\
v\xb1rm\x92%mr\xb0\x86~Fw6\x85,\
\xe8\xf7\xd3\x7f|\xffe7\xc2q\x9aN\xff\xa8\xea\x1f\
\xfb!>\x9e!yBK\x9c\x87\xe1\xe30=\xcb\xd2\
)z*Z\xeac\xbe\xc2\x03\xf6N\xfeW\xf4\xccY\
| \x9c0\xb7okw\xd8t\xb7m\xedv.\xff\
n\xdc\xcb\xd2U\xee\x17\xc5\xffl\xf3\xa2\xf8\xd5\xbfd\
\xaf\xd6\xd1\xa6y[\xb8\xc3\xe4,\xdeK\xbf\xd7->\
Rn\x16\xf7\xaaw\xa3\xc5\x19\x88E\xf2\xe4\x8ay\xf8\
s\xb2N\x028GxQW\x9b\xf5\xaa\xcaP\xd0\xce\
V\xc2\x03\x9e\xdd\xb8_\xd0\xd6I\xd9x\xe5\xe7a\xf7\
g\x819\xe1+\x9dD@\xa9 \x5c1\xf6\xadG}\
\xd1\xab\xe1\xf786\xbc\x93M\x10\xc4:\x7f\xfdJ\x89\
Q\x0a\x14\x97|B\xfd\xcfa\xc8\x88\x92Z\x1bm'\
\xc05\x01t\x04\xf86l\xd4\xb4o\x05J\xfc\x8c\xe8\
M\xf7\xee\xfe\xb7\xa6\xad\xab\x1fn\xfaEd\xfe'<\
\x9cz^\xa7\xc5\xd1\xf9\xd4\xde\xd3ex\x98\xf0\xae\x86\
j\x18\xc2a\xb0\xcen\x1e\xcd\x95\x09\x22\x85\xb4\xc6\x1c\
\xcd{\xbd\xd6I\xbb\xe4\x9c\xab\xa3\xe9\xf7d\xf2\x83h\
\x1f@\xa6\xb0\x1b\xd6\x9b\xc2M\xdd\x8b+\xab,\x1b\x84\
\xb6\xdd\xb3\x1f\xee\x9cm\x0a\xeb\xd7~\xa2\xc8K\x87\xc7\
5}\xda\xb4\xed\xf1\xdc\xbf\xab\xbc\x9c\xa2-\xb9\xba\x9f\
\x1d^vdP\xd7B\x00\x9c\x00\xb7\xe7\x10 \xfa\xd4\
PA\xd9\x05\x08\x22~W 0\x85\xe6\x0c\xea\x0c\x04\
\xc1\x89\x15\x16\xad\xf1\x12\x08\xf7e\x09\xdcz\xef\xe3\xe7\
 P\x22\xa8\xe0\x82_\x02A\xdf\x15\x08B\x13M\xcd\
(\x22\x18\x84\xc6jy\xd1\x1d\xe4}\x81 1$3\
8\x03\x01\x04aRQ\xb8\x1c\x13\xee\x0b\x05.\x09H\
\x90g(HB\xadB\x13\xb9\x94\x1c\x22\xc5\xee\x0a\x05\
\x869Y\x99QT \x06\x0dAh\xb8\x84\x82\xbd+\
\x10\x00A\xb0\xea<?\x00\x10!,\xb0\x8b\xa6\x00\xf7\
\x05\x82%x\xe2\xf6<4R\xc2\x94\x94T^\x04\xe1\
\xce2\x04:\x04g\xf6<,p@\x8f0\xe2$s\
\x9c\xc1@\xef\x0a\x06\xce|\x9f:\x82\x01\x1d\x05\x8b\xff\
\xcb\x1eqg\x81\x01\xebd#\xd8\xb9K \xa7\x06\xa6\
/&J\xb8\xb7\xf2\xd9\x12\x8a\xa9r\x5c/\x08\x90J\
],\x9f\xe1\xff\xa5\x80\xf6\x12\x1f\xe90\xdc\xb7T%\
n\xdbVu\x94n\xea\x97\xa4\xdd\xd4(?}G[\
\xc1N\xc0\xf1\xd7\x16\x81\x22T\x1a-\xad\x9c ~\xd8\
\x83)\xc9\x02\x8d\xa1\x94R\x8c\xa6\x13I\xb8\xb6\xe6\x1a\
t\xae\x87\xe3\xe3`t\x83\x22\xc7\xff\xa6\xa2\x9f\xcb\x92\
f\x99\xd4u\xf26-\xab\xd2\xfd\xafa\xe3\xe7\xb0\x01\
\xd6_\xc6R&<l\x9aX\x90A\x1aP\xc2)\xd5\
 &\x116+\x8c\xa9@x \x01\xd8$\xd2D\x82\
\x10\xa3\x89OE\xd6\xdf\x8a\x1a\x83\xa2\x88[\xb28y\
\x0e\x1d\xf3\x17-\x06\xabW\x84\x0e\x10:\x8e\xb8pB\
\x19\x160|\x12a\x87#\xe8U\xc1\xf9\x03\xb8`\x17\
a\xb9f\xfa\x96p\xd1#\x93\xf2\x99\x8c\xa2\xa4\x93\xae\
\xd0c\x18\xb4\x02\xc03\x15\xda\x084)N\xb0\x15\xba\
\xaa\x92\xfb\x88\xc1\x08eAiuK\xc0\xd8q\x88\xe2\
h#Rz`\xb0\x19\x12\xd2\x06\xd0]\x89p4\xf6\
\x09\x10k\xf5g\x1b\x8c\x15\xfe\xe6\x0d\xf8\x0d\xe1\xa2F\
\xa1\x1b\x1dIjm\xb5w$n\x88VB\x06\x11\xe2\
\xa3\x8c\xe1v\xa2\x08c\xf4sC7:\x12Z\xa5e\
7e/j\x14\x9b1\x1e3\x90\xc6\xb2\xc9\xee\x1e\x05\
sZ\x10\x81\xff<Cj\x0e\x93\x08\x88\xd2W\xf5\x02\
\x1f2\x18.\x05\x03sK\xc0\x8c\x1c\x09\xabD-\xac\
\x06\x9f\xeb\xb1*\xe2T\xeb\x003<gT\xa0\x0d\x01\
\xe1X\x08|6.\x00XA\xdcR\x09\xa4G~$\
\x18\x91\x06\xe3\xcb\xa4\xbb\x80\xa5\x9a\x8b \xb2\xc4sJ\
\xe6\xe3.\x97\xea\xb3\xe3\xae\xb6T\xe0\xdbn\xc9\x5c\xf4\
(Q\x0b\x7f|\xfe\xa6m\x17_$\xd5\x22\xc0\x1ce\
%\xef\xbc\x08\xa3\x01U\x9f\x1e`8U\x1a\xb4\xbd%\
`\xc6\x99\x1a\x83,(e\xba\xc0K\x09\xa0\xe3\xa3\x1f\
\x01\x005\x1a\xe3\xae0\x9f\x0e\x8b\x16\x1c\xbbzzK\
~t\xd2\x16\xec\xe2.\xf6d\x98\xa0YW\xd8\xf9K\
|\xa3\x03\x835\xa9Tz\xc2\x08hyU\xb7\xf9\x11\
T\xb0K\x91\x8c\xdeR\x96\xd6\xa7=\xf5.\xec2b\
0\x1f)\x9f\x8f8\x10\xa1\xb13\xc0\x0e\xc1\x80\xc2\xfc\
\xe4;\x01\x8bP}64\x9cbok\xd9\x8dAs\
\xd2i\xef\xc0\xb1\xd8\x08\xe9]\xb2\xc6 \x83\x15\x1dF\
_\x1f\x85\x95B\xc0\xb0\x9ca n\x1b\x9cY\xbc\xe8\
?\xf3G$\x1e\xde\x91\xb3\xebPO\xee\x10\xf0\xf4\x85\
\xc5\x06P\xf2\xff**'Fj\x81\xf1E\x8e\x84\xaa\
\xabM\xd9\xebyE\x83< \xe8\xcf\xc3\xcb)\xe0\xe8\
\x03\xe4\xfe\xab;\xc2\xc72\xb4\xd4\x810|\x8bgD\
\xf1\x17+\xbe\xdf\xd4\xf4(\xa1\xee>\xcc\xc6\xca\xd3\x1e\
\xb5pu7K\xc0(-\x87oPx\xccf\xfe\xcb\
\x0d\x8f\x0f\xff\x01\x9bj\x14\x0c\
\x00\x00\x0c/\
<\
?xml version=\x221.\
0\x22 encoding=\x22UTF\
-8\x22 standalone=\x22\
no\x22?>\x0a<!-- Creat\
ed with Inkscape\
 (http://www.ink\
scape.org/) -->\x0a\
\x0a<svg\x0a   xmlns:d\
c=\x22http://purl.o\
rg/dc/elements/1\
.1/\x22\x0a   xmlns:cc\
=\x22http://creativ\
ecommons.org/ns#\
\x22\x0a   xmlns:rdf=\x22\
http://www.w3.or\
g/1999/02/22-rdf\
-syntax-ns#\x22\x0a   \
xmlns:svg=\x22http:\
//www.w3.org/200\
0/svg\x22\x0a   xmlns=\
\x22http://www.w3.o\
rg/2000/svg\x22\x0a   \
xmlns:sodipodi=\x22\
http://sodipodi.\
sourceforge.net/\
DTD/sodipodi-0.d\
td\x22\x0a   xmlns:ink\
scape=\x22http://ww\
w.inkscape.org/n\
amespaces/inksca\
pe\x22\x0a   width=\x2248\
\x22\x0a   height=\x2248\x22\
\x0a   viewBox=\x220 0\
 48.000001 48.00\
0001\x22\x0a   id=\x22svg\
2\x22\x0a   version=\x221\
.1\x22\x0a   inkscape:\
version=\x220.92.2 \
(5c3e80d, 2017-0\
8-06)\x22\x0a   sodipo\
di:docname=\x22resi\
ze.svg\x22>\x0a  <defs\
\x0a     id=\x22defs4\x22\
 />\x0a  <sodipodi:\
namedview\x0a     i\
d=\x22base\x22\x0a     pa\
gecolor=\x22#ffffff\
\x22\x0a     bordercol\
or=\x22#666666\x22\x0a   \
  borderopacity=\
\x221.0\x22\x0a     inksc\
ape:pageopacity=\
\x220.0\x22\x0a     inksc\
ape:pageshadow=\x22\
2\x22\x0a     inkscape\
:zoom=\x227.9195959\
\x22\x0a     inkscape:\
cx=\x2223.580795\x22\x0a \
    inkscape:cy=\
\x2231.52398\x22\x0a     \
inkscape:documen\
t-units=\x22px\x22\x0a   \
  inkscape:curre\
nt-layer=\x22layer1\
\x22\x0a     showgrid=\
\x22false\x22\x0a     uni\
ts=\x22px\x22\x0a     ink\
scape:window-wid\
th=\x221863\x22\x0a     i\
nkscape:window-h\
eight=\x221025\x22\x0a   \
  inkscape:windo\
w-x=\x2257\x22\x0a     in\
kscape:window-y=\
\x2227\x22\x0a     inksca\
pe:window-maximi\
zed=\x221\x22 />\x0a  <me\
tadata\x0a     id=\x22\
metadata7\x22>\x0a    \
<rdf:RDF>\x0a      \
<cc:Work\x0a       \
  rdf:about=\x22\x22>\x0a\
        <dc:form\
at>image/svg+xml\
</dc:format>\x0a   \
     <dc:type\x0a  \
         rdf:res\
ource=\x22http://pu\
rl.org/dc/dcmity\
pe/StillImage\x22 /\
>\x0a        <dc:ti\
tle />\x0a      </c\
c:Work>\x0a    </rd\
f:RDF>\x0a  </metad\
ata>\x0a  <g\x0a     i\
nkscape:label=\x22C\
apa 1\x22\x0a     inks\
cape:groupmode=\x22\
layer\x22\x0a     id=\x22\
layer1\x22\x0a     tra\
nsform=\x22translat\
e(0,-1004.3622)\x22\
>\x0a    <rect\x0a    \
   style=\x22fill:n\
one;fill-opacity\
:0.84951453;stro\
ke:#999999;strok\
e-width:3.857414\
25;stroke-linejo\
in:round;stroke-\
miterlimit:4;str\
oke-dasharray:no\
ne\x22\x0a       id=\x22r\
ect4136\x22\x0a       \
width=\x2244.142586\
\x22\x0a       height=\
\x2244.142586\x22\x0a    \
   x=\x221.9287071\x22\
\x0a       y=\x221006.\
2909\x22\x0a       ry=\
\x2210.186751\x22 />\x0a \
   <g\x0a       id=\
\x22g4157\x22\x0a       t\
ransform=\x22matrix\
(0.84603094,0,0,\
0.78857975,5.043\
4872,217.0903)\x22\x0a\
       style=\x22st\
roke:#999999;fil\
l:#999999\x22>\x0a    \
  <path\x0a        \
 sodipodi:nodety\
pes=\x22ccc\x22\x0a      \
   inkscape:conn\
ector-curvature=\
\x220\x22\x0a         id=\
\x22path4138\x22\x0a     \
    d=\x22m 9.18824\
93,1043.1259 28.\
1788757,-28.9184\
 0,0\x22\x0a         s\
tyle=\x22fill:#9999\
99;fill-rule:eve\
nodd;stroke:#999\
999;stroke-width\
:4.74617958;stro\
ke-linecap:butt;\
stroke-linejoin:\
miter;stroke-mit\
erlimit:4;stroke\
-dasharray:none;\
stroke-opacity:1\
\x22 />\x0a      <path\
\x0a         sodipo\
di:nodetypes=\x22cc\
cc\x22\x0a         ink\
scape:connector-\
curvature=\x220\x22\x0a  \
       id=\x22path4\
140-3\x22\x0a         \
d=\x22m 38.871283,1\
019.5549 -6.5631\
79,-7.2978 8.297\
448,-1.8083 z\x22\x0a \
        style=\x22f\
ill:#999999;fill\
-rule:evenodd;st\
roke:#999999;str\
oke-width:1px;st\
roke-linecap:but\
t;stroke-linejoi\
n:miter;stroke-o\
pacity:1\x22 />\x0a   \
   <path\x0a       \
  sodipodi:nodet\
ypes=\x22cccc\x22\x0a    \
     inkscape:co\
nnector-curvatur\
e=\x220\x22\x0a         i\
d=\x22path4140-3-6\x22\
\x0a         d=\x22m 6\
.8693068,1038.42\
58 6.5632192,7.2\
979 -8.2974502,1\
.8083 z\x22\x0a       \
  style=\x22fill:#9\
99999;fill-rule:\
evenodd;stroke:#\
999999;stroke-wi\
dth:1px;stroke-l\
inecap:butt;stro\
ke-linejoin:mite\
r;stroke-opacity\
:1\x22 />\x0a    </g>\x0a\
  </g>\x0a</svg>\x0a\
\x00\x00\x0a\xd3\
\x00\
\x000Hx\x9c\xedZ\xdb\x8e\xdbF\x12}\xf7Wp\
\x95\x17\x1b\xcbn\xf5\xfd\x22\x8f\x1c`\x13\x04\xc8\xebn\
\x16\xfb\xb8\xa0%zFk\x89\x1cH\x1c\x8f\xed\xaf\xcf\
\xa9\xe6E\x12)a\x1dg\x03\xec\x8c\xa3\x81-\xb2\xbb\
\xba\xbb\xfat\xd5\xa9\xaa\x9e\xb9\xf9\xfe\xe3n\x9b}(\
\xf7\x87M]-g\x92\x8bYVV\xabz\xbd\xa9n\
\x97\xb3\x7f\xfe\xf2\x13\x0b\xb3\xec\xd0\x14\xd5\xba\xd8\xd6U\
\xb9\x9cU\xf5\xec\xfb7/n\xfe\xc2X\xf6\xc3\xbe,\
\x9ar\x9d=n\x9a\xbb\xec\xe7\xea\xfdaU\xdc\x97\xd9\
\xcb\xbb\xa6\xb9_\xcc\xe7\x8f\x8f\x8f|\xd35\xf2z\x7f\
;\x7f\x951\xf6\xe6\xc5\x8b\x9b\xc3\x87\xdb\x17Y\x96a\
\xdd\xea\xb0X\xaf\x96\xb3n\xc0\xfd\xc3~\x9b\x04\xd7\xab\
y\xb9-we\xd5\x1c\xe6\x92\xcb\xf9\xec(\xbe:\x8a\
\xafh\xf5\xcd\x87rU\xefvuuH#\xab\xc3w\
'\xc2\xfb\xf5\xbbA\x9a\xb4y\xd4IH\xc6\x18\xe7B\
\xcd\x95b\x90`\x87OUS|d\xe7C\xa1\xe3\xa5\
\xa1J\x081G\xdfQ\xf2\xcb\xa4\x16\x07\x00z\x8f\x7f\
\x83x\xdf\xc0\x0f\xf5\xc3~U\xbe\xc3\xb8\x92We3\
\xff\xf1\x97\x1f\x87N&\xf8\xbaY\x9fL\xd3\xe3y\xb6\
\xea\x19\xc8U\xb1+\x0f\xf7\xc5\xaa<\xcc\xfb\xf64\xfe\
q\xb3n\xee\x963\x13\xd2\xdb]\xb9\xb9\xbdk\x86\xd7\
\x0f\x9b\xf2\xf1o\xf5\xc7\xe5Ld\x223\x81\x0b\xfa\xc8\
\xe3S\x12\xda\xac\x973lJ\xb5#\x8e\x06\xd3\xf5v\
\x8b-\x86\x1e\xc1\xa3\xe2:{\xa9\x8c\xb0\xd6\xb8<S\
B\x06&4\x93\xf2U\x1a\xd2\xefr\xb1\xaeW\xa4\xf6\
rv\xbb\xdf\xac\xff\xbdY\xd5\x15'\xf4\xde@\xe8f\
]\xbe;\x90p\xbb>\xbd\x99Y6O]\xc3x\x1a\
\xbc\xa6=\x1c\x05\xdf\x16\x87v\xdfYv_\xdc\xc2F\
\xb6\xf5~9\xfb\xee]\xfat\x1do\xeb\xfd\xba\xdc\xf7\
].}\xce\xbaj\xe0\xb8i>\xb5^\xd1\xcd\xddo\
\x93f\x1d\xfa\xc5\xe5\xfe\xc3]\xb1\xae\x1f\x9735\xee\
\xfc\x5c\xd7;\x9a\xd5\x1a\x15]\x08\xe3\xee\x15N\x82I\
c\xb9\xb0\xd2\xcaI/)$%\xd7\xd1y3\xee\x04\
\x94\x0f\xe48\xec\xa1\xda40\xce\xfb\x8f\x93\xe1\x0f\xfb\
=\x09l\x8bO%\xf6\x9d\xbe\xfa5\x0ew\xf5#\x9d\
\xc1r\xf6\xae\xd8\x0e\x00^\x9d\xeaqSa\x7f\xac3\
-\x19\xa6\xbav\x12\xbd\xb9I\xa1\xec\x15\x11l\xd9\xc5\
+}\xd8\xb0\xf2W\xfav\xc5\xc7\xcdn\xf3\xb9\x84\xce\
\xb27\x8c]\xd9\x14\xeb\xa2)\x8e\xe6\xd0\xb7\xf8dT\
\x10\x81\xe3/\xfe\xfe\xe3O\xed\x1b\xdeW\xab\xc5\xbf\xea\
\xfd\xfb\xee\x15\x1f\x12(\xde\xd6\x0f\xd0z\xf6fh\xbe\
Y\xaf\x16p\xd5]\xd1\xbc\xd9\xecp\xc2\xe4\xe5\x7f\x85\
k\xde\xcc\x8f\x1dg\xc2\xcd\xa7\xfb\xf28i;\xed\xbe\
l}\xfe\x22\xf1\xadW\xbb\x0d\x0d\x9a\xff\xa3\xd9l\xb7\
?\xd3\x22\xdd\xb6N&\xdd4\xdb\xf2\xd8x3\xef\xb4\
\xef\xf66?\xd9\xdc\xcd\xbc\xdfzz\xbb\x1d\x81\xb8-\
\xde\x96\xdb\xe5\xec\x87\xe2\xbe\xc8&\x87w\xbb\xaf\x1f\xee\
w\xf5\xba\xec\xacdv\xc4\xf3\xccj\x9a}Q\x1dh\
\xf3\xcbYz\xdc\x22(\xbc\x149\x93B\x18\xae\x9dR\
\xafz\xd4o\xfbm\xd0\x1c\xb7\xc1\xebY\xdfpq\x0e\
\x16\xe0\x02Rh\x90\x073\x8a\x07\x13Bx5\x0c9\
4\x9f\xb6P\xed\xd0\xec\xeb\xf7\xe5\xe2\xbb\x98>\xaf\xdb\
W\xd6\xb9\xe6B\x0e\x87wSn\xb7\x9b\xfb\xc3\xc9i\
\xecaW\x86\xc7\xe0\x9d\x0a\xb3\x93f\x98\xa2\xe5\xd2H\
\x11\xb5?iO~'\xa4\xe2\xd2\xc6\xb3v\xc8G\x10\
epZ\x9d\xb6\xd3\x1e\xef\x8b\xe6\xceH\xedN\x9a;\
\xb5\x07\x05_\xbf\xc39\x0f\xea\xd3\x0b;\xefc\xfb\x87\
m\xb9(?\x94U\xbd^\xbf\xbe\xbc\xdd\xe4\x82\x0b\xdb\
\xbfn7U\xf9\x9fzS-p\x82U?\x86\xc1\xb0\
\xca\xfd\x16\xde\xd2,L\xdf\xb6.@R\xfb}\xf1i\
Q!\xbaO\xd1;\xb1\xb2+\xf8\x99 \x1d\x90\x1a\x01\
h\xb8\x93\xd1\x0ak&\x00\x1a\xc1\xa5\xf2z\x04 \xb5\
Jo\x8c\xbe\x02 \x93\xcf\x16B\xcbM\x8cZ\x08;\
\xb1A\xe7@\xa9b\x0a!\xc8\xdf\xd8\xe0F\x10\xea\xc8\
\x9d\x95J\xb9k\x10>_+t\x5c{g\x83\x8f#\
\x08\x1d\xb7^\xf9(\xe5\xd4\x0a\x15\xb7R\x8f\xad\xd0s\
g\x9c\xf0\xd7\xadP=[\x08-\xf7\xc6*sf=\
\xad\x15F\x85\xc4-\xda\x09\x84\xca\x03u\x13G\x10*\
\xcb\xc1\xd2\xc1]\xb5B\xf1\x1c \xa4\xdd\x9clcH\
D\x11,)x#[Z\xadN\x01\x18\x92\xaf\xba\xaa\
\xcaUS\xef\x19\xd2\xb0\x0fE\xf3\xb0\xc7\xf6\xc5E\xa8\
\xfc\xa9\xd1R\x0a\x93\xc1\xed]P\x22\x9a\x1c11\x82\
/\x83\xcc\xa4F\x92\xa8\x90&\xe6,r\xe3\xc5\x05\x03\
\x9db\xfa\xc5 \x9aS\x10\xb1\x81s\x0c\x07d\x13\x98\
O\x08Y=AV\xc0h\x95%`\x01(\xf2jE\
\xc0\xba\xa8\xa5\x8a9\x22\xbe\x07\xbb\xfc\x09\xec\x7f\x07\xd6\
\x8e\x81\xd5\x81k\xe1\xb4V-\xb2\xce{\x991\xaa]\
\xb4\xb4V\x12\xb4\xd6\xc9\x0bq\xe9Oh\xc7\xd0\xfa1\
\xb4\xa0_\x85<J\xc2@\x85\x8a<\x22\xd9&\x86\xd0\
>h\xe3s\xd0\x83\x0a\xf1I\x03{3\xbf\xbd\x5c6\
\x84#\x16\x17\xcb\x06$\xee\xc8\xea\x9d\xd1y@\xf9\x10\
\x831\xaffW\xce\xea\xb7\x1fEP\x13^V\x8a\xe3\
 \xa2\x0ay*\x0f\xac\x886\xd3\xa8\x12\x8c\xb26\xd7\
 \x0f\xa53\xc9\x83\x8e1\xc8\x1cqV\x18\x93y\x14\
\x0c\xd1X\x9f3\x0c\xf0\xeeBv\x9bN*!t\xfd\
\x98\x82\xa0\x9f\xf3c\xd2_vLO*\x1a\x06\x15\xbf\
\x04t\xe6Q\xb6\x19\x11\x03\xcc\xdfDq\xcd\xfc\x9f\x22\
\xa8_\x81\x99\x9e\x1a*(\xc3\xa1\x5c\xa38\xa7$\xf9\
\x88\xcaP}h\x85\xc25\xf7<\x8222\x86\x8aX\
8'\x03,U\xb8\xb3\xe2\xe3[\xc4p\x92*\x5c\xb2\
;\xc9\xa5\x8eB\xc3\x9958Y\xe8g\x84\xd9eg\
\xfe\xbd\xee\xac'\x89\x82\xb4\x08X1h\xb2LE\xb7\
\xce&\xa3\xdc\xc0[\x13a\x99A)\x93\xad2&\x10\
\xdeT\xf0\x1e\xac\xaa\xb5\xf7\x99\x80\x89Z\x97+\x0e{\
uH,('v)\xd9\x00\xf1\xebl\x0b\x1a\xb6\xc2\
\xa0\x04\x841[%\xae\xe5pO\xf1`\xbe\x06\xf5I\
\x0e\xa1\x91\xda\x02\xea(\xdb\x1c\x22\xc4\x989\xb2`\x84\
\xd8\xdcpa\x82&\x04\xa3\xa2\x8c\xc2q\xe1\xd57\x8e\
\xa0\x99\xd2\x81\xe7>\xea\xe0\xc9p5\xa8\xd3F\x03\xab\
D&\xac4\x19\x9dQ\xe19A\xf6G\x84v3\xe5\
\x02\xc1\xe9*U%\xab\x0c\x88O\xc0\x949n\xa4\xf1\
\x08\xea@\xd5\x89\xf8\x8d\xc7v3-\x07\x92\xb1\xc9\xc4\
\xa0\x12N\x8cG\xc4\xf6\x80d\x93\xee\x06\x14G\xc8B\
QK\xd4\xaa]@\xcd\xc0\xbd\x09\xd7.\x0b\xbe\x15\x10\
'I%<X\xa1\xd6O\x96'\x91C\xe2\xff\x8c\xaa\
\xd6\x18m\x00!Z\x11Lf8\x8a.\x1dL\x0e7\
GN\xff\x8c \xfccb\xbd\xfd\x92X\x8f\xac\xd3D\
\x8f\xc2\x0da\x06\xfe\x9e\xc1`\xa54\xa8\xdc(/\x05\
T\x92$\x12\xee\xd2!\xb9\x87\x85\xfbg\x04\xfc\xd7\xa0\
z\xe9> \xba(\xc9t\x0d\xf2\xa4\xe0#\xf2R\xa5\
\x9c\xb3\x0au\xb0\x0c\xe1\xda5\xcbSD\xec\x8a\xa9\xfe\
\xceH\xe4\xd9$\xbe\x1bNe\xbb\xb2\xed\xfd\x95\xb4\xda\
fd\xaa\xd6\x06\xe3\x10\x89\x82u\x81R&m\x8c\x12\
\x06\x89h\x8c\xe2\x09\xd7\xf2W\xaf\x5c\xfc\xf1\x17\xa4W\
/\x92\x8e7\xf5\x97o\x91\x8e\xcb]\xb8\xbd\xd9\x15\xcd\
~\xf3\xf1\xa5@\xce\x1fMD\xfa)\xe8gx\x83q\
;k\x82P\xb9\xa1\x18f\x00\xff\xf5;\x9d\xee<\xd5\
\x84w\x1cG=\xe1\x14\xd5\x07QrcP\x22\xac\xa8\
j\x80\x87\x08\xa2\x1c\xc4E,\x09\xf6\xd1<\x82\x9d\xb8\
5(!\x98DP\xb0\xf4g0:g\x81k\x891\
\xf4\xcb\xca\x1c\xd5\x88\xb1\xf4\x18\xe81\xaa\xee\x11\x13g\
\x22u\x06*X\x84O\xbd\x96\x9e\xfb\xeem\x869\xa5\
\x18\xe6T\xa4\x87$QIU\x0fVFa\x8e\xfc;\
\xb4O(v$\x14t\x14\xc6]\xa4oG+Ps\
\x96\x9a\xa8\x81\xa5f\xd6\xc9\xb2T\x14E\xfa&\xd9\xa4\
L\xd2%\xa9\x924i\x15a'\x9at\x8a\xb0\xa4\x09\
Ks[R\x18*\xb0\xa4M\xf7\x8c5\x04\x1d\xac\xa7\
\x85,oO\xd9w\x0b\x1e\xdfi\xe1\xa4cR\xd1e\
\xa7=\xc06`\xb4\x06\x93\x13<\xfdv\xdb\x05\x12<\
=\xe4\x1d\xe2\x84\x1emC\xa5'\x87G*\xf8\xd2\xa3\
$\x09Z\x0a\x09\x0eU\x82\xe4\x96\x9ar\x1f\xdd>Q\
\x07\xd6J\x0d\x04\x97\xec\x1eS\xa7hk\xf6\x84\xa3\xec\
\xda\xd8I\xff\xe7\xaf\xa2\x94\xff\x81\x8f\x5c\x0f\x16\xbdy\
O\xc9\x8a![\xb62\x86@t\xe5\x02\xf7.\x84d\
\xadQXA@)%\x15E\x05\xe3\x02\x0c\x18[\xf6\
\x9a\xde)\xcd\xb6T\xf0yCE3\xc5g'\x95\xa6\
`\x8cS\xb0i\x0a\x15\x9d@\x0b\x95\xe0\xd6Q\x8d\xe3\
\xb4\x8c6\x1d^Tt\x05\x22P\xf4\xb4\x9e\xe0\x09W\
D \xeb4\xbdcn2\xbe\x80:\xb3u\x06+\x90\
\xd0C\x0bL`\xda\x19BL\xa6\xc81\xd2#\x07`\
\x94\xc0\xaa`Zc4FX\x1c3\xcd\xe4L:)\
\x94\x01R%k\x11F\x072H\xafS\xcdJy\xae\
\xc4\x96\xd0\x131\x91h\x0f\xdcF\x9d\xe6\xa1\x8c\x22\xa7\
BB\xa9\x16\x05\xef\x91\xc6\x91O\xc7\xd6JQ\xe9\x92\
\xca\xed\xe0,\x09J\x0d\xe0\xda\xc16K\x92B\xab\x04\
\xa5\xf1!\xeb\x96\x81}\xb5\xa3\xb3v\x15\x9a\xb6\x1dM\
\xeb\xa6ul\x9e\x06\xa7+\x0c\x02\x86\x8c9\x01\xc3\x06\
dDw\x899 \x93\x80\x91t =0-.-\
[\xb4\xb0\x90T\x82\xa5G\x85xKP\x0a\xd5\x81\x12\
\xf2\x84\x89\x1et\xd5.\xef0iw%e\x1c)K\
\xbfM\xec\xb4\xedA!NJ\xa0\xb0\x1e\x15\x99w\xa0\
\xb0\x0e\x15l\x89u\xb0\x90\x11\x18\xe1\xa3J-\x8e\xca\
7|G#]\xa0\x03\xf1\xc1\xa7\x16\xa9\xf0D\x7f\xda\
\x83Q0!\xb2\x80\xd6\xf2PM\xb3dz\xba\xf5|\
\xe4\xe4!\x10bZS\xaeH\x1cm\x9d#\xc4\x9c\x11\
\x89\x06\xa2\xf52\xb1w\x14\xc4\x0b \xae\xe0\xa2\xa62\
'J\xafIk\xd4\xe31\xb1\x9e\xf6&)--j\
\xf4dDp\x8a\xa4\x0f\x89\x02\x8c4\xdaS\x03\x89\xca\
d2\x90\xccZA\x9cu;Xg\xc3*\xdd\xd8l\
X\xa5\x1b\x9bu\xab\xf8\xbc[\xe4\xff\x90M~s\x9a\
\x04\xde9\xfbK\x92\x81y\x84\x93\x8eB\x96\x88\x08\xd6\
\xda\xbb\x13\xe6ac\xeaa#\xeea\x17\xc8\x87M\xd9\
gD>#\xee\x19Q\xcf9\xf3\xb8\x90\x9f3\x8f\x09\
\xf9\x98y\xe0,\x97\x98gB<\x13\xde\x99\xd0\xce\x98\
u\xc6\xa43\xe6\x9cs\xca\x19|\xabw\xad\xc1\xb3z\
'\xeex\xedH8\x1d\xaf\x9d\x12\x0e\x1b3\x0e\x9bP\
\x0e\x9bp\x0e\x9b\x92\x0e\x9b\xb0\x8e\x1f\xc8x\xc4\xc5'\
\xac\xc3F\xb4\x83\xec\x98\x9d\xf3\x8e\x1a\xebl\xe5\xa0s\
\x07\x8c\x19\xc8\x98\x0d\xd0\xb0\x11\xef\x98|\xc2:\x13\xd2\
\x99p\xce%\xca\x990N\x0av\xa7\x8c\x93\x82\xd0\x19\
\xe5\xb4\x01\xef\x9cs\xd8\x84t\xd8\x84u&\xa43\xe6\
\x9c\x11\xe5\x8c\x19gL8\xe7|\xe3\x06V\xfb\xfc\xc7\
\xd0HW'\xa4\xaf\x1b\xfa\x1b\xd87/~\x05,n\
NV\
\x00\x00\x09,\
<\
?xml version=\x221.\
0\x22 encoding=\x22UTF\
-8\x22 standalone=\x22\
no\x22?>\x0a<!-- Creat\
ed with Inkscape\
 (http://www.ink\
scape.org/) -->\x0a\
\x0a<svg\x0a   xmlns:d\
c=\x22http://purl.o\
rg/dc/elements/1\
.1/\x22\x0a   xmlns:cc\
=\x22http://creativ\
ecommons.org/ns#\
\x22\x0a   xmlns:rdf=\x22\
http://www.w3.or\
g/1999/02/22-rdf\
-syntax-ns#\x22\x0a   \
xmlns:svg=\x22http:\
//www.w3.org/200\
0/svg\x22\x0a   xmlns=\
\x22http://www.w3.o\
rg/2000/svg\x22\x0a   \
xmlns:sodipodi=\x22\
http://sodipodi.\
sourceforge.net/\
DTD/sodipodi-0.d\
td\x22\x0a   xmlns:ink\
scape=\x22http://ww\
w.inkscape.org/n\
amespaces/inksca\
pe\x22\x0a   version=\x22\
1.1\x22\x0a   id=\x22svg2\
\x22\x0a   width=\x22192\x22\
\x0a   height=\x22192\x22\
\x0a   viewBox=\x220 0\
 192 192\x22\x0a   sod\
ipodi:docname=\x22p\
lus (gray).svg\x22\x0a\
   inkscape:vers\
ion=\x220.92.4 (unk\
nown)\x22>\x0a  <metad\
ata\x0a     id=\x22met\
adata8\x22>\x0a    <rd\
f:RDF>\x0a      <cc\
:Work\x0a         r\
df:about=\x22\x22>\x0a   \
     <dc:format>\
image/svg+xml</d\
c:format>\x0a      \
  <dc:type\x0a     \
      rdf:resour\
ce=\x22http://purl.\
org/dc/dcmitype/\
StillImage\x22 />\x0a \
       <dc:title\
 />\x0a      </cc:W\
ork>\x0a    </rdf:R\
DF>\x0a  </metadata\
>\x0a  <defs\x0a     i\
d=\x22defs6\x22 />\x0a  <\
sodipodi:namedvi\
ew\x0a     pagecolo\
r=\x22#ffffff\x22\x0a    \
 bordercolor=\x22#6\
66666\x22\x0a     bord\
eropacity=\x221\x22\x0a  \
   objecttoleran\
ce=\x2210\x22\x0a     gri\
dtolerance=\x2210\x22\x0a\
     guidetolera\
nce=\x2210\x22\x0a     in\
kscape:pageopaci\
ty=\x220\x22\x0a     inks\
cape:pageshadow=\
\x222\x22\x0a     inkscap\
e:window-width=\x22\
1863\x22\x0a     inksc\
ape:window-heigh\
t=\x221025\x22\x0a     id\
=\x22namedview4\x22\x0a  \
   showgrid=\x22fal\
se\x22\x0a     inkscap\
e:zoom=\x221.738304\
2\x22\x0a     inkscape\
:cx=\x22133.79315\x22\x0a\
     inkscape:cy\
=\x2271.379447\x22\x0a   \
  inkscape:windo\
w-x=\x2257\x22\x0a     in\
kscape:window-y=\
\x2227\x22\x0a     inksca\
pe:window-maximi\
zed=\x221\x22\x0a     ink\
scape:current-la\
yer=\x22svg2\x22 />\x0a  \
<g\x0a     id=\x22g831\
\x22\x0a     style=\x22st\
roke:#aeaeae;str\
oke-linecap:roun\
d;stroke-opacity\
:1;fill:#aeaeae;\
fill-opacity:1\x22>\
\x0a    <path\x0a     \
  sodipodi:nodet\
ypes=\x22cc\x22\x0a      \
 inkscape:connec\
tor-curvature=\x220\
\x22\x0a       id=\x22pat\
h812\x22\x0a       d=\x22\
m 95.091973,56.1\
72243 0.09447,79\
.795097\x22\x0a       \
style=\x22fill:#aea\
eae;fill-rule:ev\
enodd;stroke:#ae\
aeae;stroke-widt\
h:16;stroke-line\
cap:round;stroke\
-linejoin:miter;\
stroke-miterlimi\
t:4;stroke-dasha\
rray:none;stroke\
-opacity:1;fill-\
opacity:1\x22 />\x0a  \
  <path\x0a       s\
odipodi:nodetype\
s=\x22cc\x22\x0a       in\
kscape:connector\
-curvature=\x220\x22\x0a \
      id=\x22path81\
2-3\x22\x0a       d=\x22m\
 134.96147,95.59\
4487 -79.795108,\
0.0945\x22\x0a       s\
tyle=\x22fill:#aeae\
ae;fill-rule:eve\
nodd;stroke:#aea\
eae;stroke-width\
:16;stroke-linec\
ap:round;stroke-\
linejoin:miter;s\
troke-miterlimit\
:4;stroke-dashar\
ray:none;stroke-\
opacity:1;fill-o\
pacity:1\x22 />\x0a  <\
/g>\x0a</svg>\x0a\
\x00\x00\x07\x95\
<\
?xml version=\x221.\
0\x22 encoding=\x22UTF\
-8\x22 standalone=\x22\
no\x22?>\x0a<!-- Creat\
ed with Inkscape\
 (http://www.ink\
scape.org/) -->\x0a\
\x0a<svg\x0a   xmlns:d\
c=\x22http://purl.o\
rg/dc/elements/1\
.1/\x22\x0a   xmlns:cc\
=\x22http://creativ\
ecommons.org/ns#\
\x22\x0a   xmlns:rdf=\x22\
http://www.w3.or\
g/1999/02/22-rdf\
-syntax-ns#\x22\x0a   \
xmlns:svg=\x22http:\
//www.w3.org/200\
0/svg\x22\x0a   xmlns=\
\x22http://www.w3.o\
rg/2000/svg\x22\x0a   \
xmlns:sodipodi=\x22\
http://sodipodi.\
sourceforge.net/\
DTD/sodipodi-0.d\
td\x22\x0a   xmlns:ink\
scape=\x22http://ww\
w.inkscape.org/n\
amespaces/inksca\
pe\x22\x0a   version=\x22\
1.1\x22\x0a   id=\x22svg2\
\x22\x0a   width=\x22192\x22\
\x0a   height=\x22192\x22\
\x0a   viewBox=\x220 0\
 192 192\x22\x0a   sod\
ipodi:docname=\x22m\
inus (gray).svg\x22\
\x0a   inkscape:ver\
sion=\x220.92.4 (un\
known)\x22>\x0a  <meta\
data\x0a     id=\x22me\
tadata8\x22>\x0a    <r\
df:RDF>\x0a      <c\
c:Work\x0a         \
rdf:about=\x22\x22>\x0a  \
      <dc:format\
>image/svg+xml</\
dc:format>\x0a     \
   <dc:type\x0a    \
       rdf:resou\
rce=\x22http://purl\
.org/dc/dcmitype\
/StillImage\x22 />\x0a\
        <dc:titl\
e />\x0a      </cc:\
Work>\x0a    </rdf:\
RDF>\x0a  </metadat\
a>\x0a  <defs\x0a     \
id=\x22defs6\x22 />\x0a  \
<sodipodi:namedv\
iew\x0a     pagecol\
or=\x22#ffffff\x22\x0a   \
  bordercolor=\x22#\
666666\x22\x0a     bor\
deropacity=\x221\x22\x0a \
    objecttolera\
nce=\x2210\x22\x0a     gr\
idtolerance=\x2210\x22\
\x0a     guidetoler\
ance=\x2210\x22\x0a     i\
nkscape:pageopac\
ity=\x220\x22\x0a     ink\
scape:pageshadow\
=\x222\x22\x0a     inksca\
pe:window-width=\
\x221863\x22\x0a     inks\
cape:window-heig\
ht=\x221025\x22\x0a     i\
d=\x22namedview4\x22\x0a \
    showgrid=\x22fa\
lse\x22\x0a     inksca\
pe:zoom=\x221.73830\
42\x22\x0a     inkscap\
e:cx=\x22-319.39807\
\x22\x0a     inkscape:\
cy=\x22209.44504\x22\x0a \
    inkscape:win\
dow-x=\x2257\x22\x0a     \
inkscape:window-\
y=\x2227\x22\x0a     inks\
cape:window-maxi\
mized=\x221\x22\x0a     i\
nkscape:current-\
layer=\x22g831\x22 />\x0a\
  <g\x0a     id=\x22g8\
31\x22\x0a     style=\x22\
stroke:#b3b3b3;s\
troke-linecap:ro\
und\x22>\x0a    <path\x0a\
       sodipodi:\
nodetypes=\x22cc\x22\x0a \
      inkscape:c\
onnector-curvatu\
re=\x220\x22\x0a       id\
=\x22path812-3\x22\x0a   \
    d=\x22m 134.961\
47,95.594487 -79\
.795108,0.0945\x22\x0a\
       style=\x22fi\
ll:#37c8ab;fill-\
rule:evenodd;str\
oke:#aeaeae;stro\
ke-width:16;stro\
ke-linecap:round\
;stroke-linejoin\
:miter;stroke-mi\
terlimit:4;strok\
e-dasharray:none\
;stroke-opacity:\
1\x22 />\x0a  </g>\x0a</s\
vg>\x0a\
"

qt_resource_name = b"\
\x00\x05\
\x00O\xa6S\
\x00I\
\x00c\x00o\x00n\x00s\
\x00\x05\
\x00o\xa6S\
\x00i\
\x00c\x00o\x00n\x00s\
\x00\x14\
\x0b3\xdd\x87\
\x00a\
\x00u\x00t\x00o\x00m\x00a\x00t\x00i\x00c\x00_\x00l\x00a\x00y\x00o\x00u\x00t\x00.\
\x00s\x00v\x00g\
\x00\x0a\
\x01\x0bC\x87\
\x00r\
\x00e\x00s\x00i\x00z\x00e\x00.\x00s\x00v\x00g\
\x00\x0d\
\x07~\x10\x87\
\x00g\
\x00r\x00i\x00d\x00_\x00i\x00c\x00o\x00n\x00.\x00s\x00v\x00g\
\x00\x0f\
\x00 ['\
\x00p\
\x00l\x00u\x00s\x00 \x00(\x00g\x00r\x00a\x00y\x00)\x00.\x00s\x00v\x00g\
\x00\x10\
\x00\x04\x86\xa7\
\x00m\
\x00i\x00n\x00u\x00s\x00 \x00(\x00g\x00r\x00a\x00y\x00)\x00.\x00s\x00v\x00g\
"

qt_resource_struct = b"\
\x00\x00\x00\x00\x00\x02\x00\x00\x00\x01\x00\x00\x00\x01\
\x00\x00\x00\x00\x00\x00\x00\x00\
\x00\x00\x00\x00\x00\x02\x00\x00\x00\x01\x00\x00\x00\x02\
\x00\x00\x00\x00\x00\x00\x00\x00\
\x00\x00\x00\x10\x00\x02\x00\x00\x00\x05\x00\x00\x00\x03\
\x00\x00\x00\x00\x00\x00\x00\x00\
\x00\x00\x00\xac\x00\x00\x00\x00\x00\x01\x00\x00'h\
\x00\x00\x01s\xa14\xaf\xfc\
\x00\x00\x00\x88\x00\x00\x00\x00\x00\x01\x00\x00\x1e8\
\x00\x00\x01s\xa14\xb0\xc8\
\x00\x00\x00N\x00\x00\x00\x00\x00\x01\x00\x00\x07.\
\x00\x00\x01s\xa14\xb0 \
\x00\x00\x00h\x00\x01\x00\x00\x00\x01\x00\x00\x13a\
\x00\x00\x01s\xa14\xb0\xa0\
\x00\x00\x00 \x00\x01\x00\x00\x00\x01\x00\x00\x00\x00\
\x00\x00\x01s\xa14\xb0\x0c\
"

def qInitResources():
    QtCore.qRegisterResourceData(0x03, qt_resource_struct, qt_resource_name, qt_resource_data)

def qCleanupResources():
    QtCore.qUnregisterResourceData(0x03, qt_resource_struct, qt_resource_name, qt_resource_data)

qInitResources()
