# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'gui.ui',
# licensing of 'gui.ui' applies.
#
# Created: Tue Jun 25 21:00:49 2019
#      by: pyside2-uic  running on PySide2 5.12.3
#
# WARNING! All changes made in this file will be lost!

from PySide2 import QtCore, QtGui, QtWidgets

class Ui_Dialog(object):
    def setupUi(self, Dialog):
        Dialog.setObjectName("Dialog")
        Dialog.resize(941, 590)
        self.verticalLayout = QtWidgets.QVBoxLayout(Dialog)
        self.verticalLayout.setContentsMargins(1, 1, 1, 1)
        self.verticalLayout.setObjectName("verticalLayout")
        self.splitter_3 = QtWidgets.QSplitter(Dialog)
        self.splitter_3.setOrientation(QtCore.Qt.Horizontal)
        self.splitter_3.setObjectName("splitter_3")
        self.frame_8 = QtWidgets.QFrame(self.splitter_3)
        self.frame_8.setFrameShape(QtWidgets.QFrame.NoFrame)
        self.frame_8.setFrameShadow(QtWidgets.QFrame.Raised)
        self.frame_8.setObjectName("frame_8")
        self.verticalLayout_7 = QtWidgets.QVBoxLayout(self.frame_8)
        self.verticalLayout_7.setContentsMargins(-1, 9, -1, -1)
        self.verticalLayout_7.setObjectName("verticalLayout_7")
        self.toolBox = QtWidgets.QToolBox(self.frame_8)
        self.toolBox.setObjectName("toolBox")
        self.page_3 = QtWidgets.QWidget()
        self.page_3.setObjectName("page_3")
        self.verticalLayout_3 = QtWidgets.QVBoxLayout(self.page_3)
        self.verticalLayout_3.setContentsMargins(0, -1, 0, -1)
        self.verticalLayout_3.setObjectName("verticalLayout_3")
        self.frame_2 = QtWidgets.QFrame(self.page_3)
        self.frame_2.setMinimumSize(QtCore.QSize(300, 0))
        self.frame_2.setFrameShape(QtWidgets.QFrame.StyledPanel)
        self.frame_2.setFrameShadow(QtWidgets.QFrame.Raised)
        self.frame_2.setObjectName("frame_2")
        self.verticalLayout_4 = QtWidgets.QVBoxLayout(self.frame_2)
        self.verticalLayout_4.setObjectName("verticalLayout_4")
        self.frame_3 = QtWidgets.QFrame(self.frame_2)
        self.frame_3.setFrameShape(QtWidgets.QFrame.NoFrame)
        self.frame_3.setFrameShadow(QtWidgets.QFrame.Plain)
        self.frame_3.setObjectName("frame_3")
        self.horizontalLayout_3 = QtWidgets.QHBoxLayout(self.frame_3)
        self.horizontalLayout_3.setObjectName("horizontalLayout_3")
        self.label = QtWidgets.QLabel(self.frame_3)
        self.label.setMaximumSize(QtCore.QSize(160, 16777215))
        self.label.setObjectName("label")
        self.horizontalLayout_3.addWidget(self.label)
        self.night_valley_timeEdit = QtWidgets.QTimeEdit(self.frame_3)
        self.night_valley_timeEdit.setTime(QtCore.QTime(3, 0, 0))
        self.night_valley_timeEdit.setObjectName("night_valley_timeEdit")
        self.horizontalLayout_3.addWidget(self.night_valley_timeEdit)
        self.night_valley_doubleSpinBox = QtWidgets.QDoubleSpinBox(self.frame_3)
        self.night_valley_doubleSpinBox.setProperty("value", 4.0)
        self.night_valley_doubleSpinBox.setObjectName("night_valley_doubleSpinBox")
        self.horizontalLayout_3.addWidget(self.night_valley_doubleSpinBox)
        self.verticalLayout_4.addWidget(self.frame_3)
        self.frame_4 = QtWidgets.QFrame(self.frame_2)
        self.frame_4.setFrameShape(QtWidgets.QFrame.NoFrame)
        self.frame_4.setFrameShadow(QtWidgets.QFrame.Plain)
        self.frame_4.setObjectName("frame_4")
        self.horizontalLayout_4 = QtWidgets.QHBoxLayout(self.frame_4)
        self.horizontalLayout_4.setObjectName("horizontalLayout_4")
        self.label_2 = QtWidgets.QLabel(self.frame_4)
        self.label_2.setMaximumSize(QtCore.QSize(160, 16777215))
        self.label_2.setObjectName("label_2")
        self.horizontalLayout_4.addWidget(self.label_2)
        self.morning_peak_timeEdit = QtWidgets.QTimeEdit(self.frame_4)
        self.morning_peak_timeEdit.setTime(QtCore.QTime(10, 0, 0))
        self.morning_peak_timeEdit.setObjectName("morning_peak_timeEdit")
        self.horizontalLayout_4.addWidget(self.morning_peak_timeEdit)
        self.morning_peak_doubleSpinBox = QtWidgets.QDoubleSpinBox(self.frame_4)
        self.morning_peak_doubleSpinBox.setProperty("value", 12.0)
        self.morning_peak_doubleSpinBox.setObjectName("morning_peak_doubleSpinBox")
        self.horizontalLayout_4.addWidget(self.morning_peak_doubleSpinBox)
        self.verticalLayout_4.addWidget(self.frame_4)
        self.frame_5 = QtWidgets.QFrame(self.frame_2)
        self.frame_5.setFrameShape(QtWidgets.QFrame.NoFrame)
        self.frame_5.setFrameShadow(QtWidgets.QFrame.Plain)
        self.frame_5.setObjectName("frame_5")
        self.horizontalLayout_6 = QtWidgets.QHBoxLayout(self.frame_5)
        self.horizontalLayout_6.setObjectName("horizontalLayout_6")
        self.label_4 = QtWidgets.QLabel(self.frame_5)
        self.label_4.setMaximumSize(QtCore.QSize(160, 16777215))
        self.label_4.setObjectName("label_4")
        self.horizontalLayout_6.addWidget(self.label_4)
        self.afternoon_valley_timeEdit = QtWidgets.QTimeEdit(self.frame_5)
        self.afternoon_valley_timeEdit.setTime(QtCore.QTime(15, 0, 0))
        self.afternoon_valley_timeEdit.setObjectName("afternoon_valley_timeEdit")
        self.horizontalLayout_6.addWidget(self.afternoon_valley_timeEdit)
        self.afternoon_valley_doubleSpinBox = QtWidgets.QDoubleSpinBox(self.frame_5)
        self.afternoon_valley_doubleSpinBox.setProperty("value", 8.0)
        self.afternoon_valley_doubleSpinBox.setObjectName("afternoon_valley_doubleSpinBox")
        self.horizontalLayout_6.addWidget(self.afternoon_valley_doubleSpinBox)
        self.verticalLayout_4.addWidget(self.frame_5)
        self.frame_6 = QtWidgets.QFrame(self.frame_2)
        self.frame_6.setFrameShape(QtWidgets.QFrame.NoFrame)
        self.frame_6.setFrameShadow(QtWidgets.QFrame.Plain)
        self.frame_6.setObjectName("frame_6")
        self.horizontalLayout_7 = QtWidgets.QHBoxLayout(self.frame_6)
        self.horizontalLayout_7.setObjectName("horizontalLayout_7")
        self.label_5 = QtWidgets.QLabel(self.frame_6)
        self.label_5.setMaximumSize(QtCore.QSize(160, 16777215))
        self.label_5.setObjectName("label_5")
        self.horizontalLayout_7.addWidget(self.label_5)
        self.evening_peak_timeEdit = QtWidgets.QTimeEdit(self.frame_6)
        self.evening_peak_timeEdit.setTime(QtCore.QTime(20, 0, 0))
        self.evening_peak_timeEdit.setObjectName("evening_peak_timeEdit")
        self.horizontalLayout_7.addWidget(self.evening_peak_timeEdit)
        self.evening_peak_doubleSpinBox = QtWidgets.QDoubleSpinBox(self.frame_6)
        self.evening_peak_doubleSpinBox.setProperty("value", 16.0)
        self.evening_peak_doubleSpinBox.setObjectName("evening_peak_doubleSpinBox")
        self.horizontalLayout_7.addWidget(self.evening_peak_doubleSpinBox)
        self.verticalLayout_4.addWidget(self.frame_6)
        self.frame_7 = QtWidgets.QFrame(self.frame_2)
        self.frame_7.setFrameShape(QtWidgets.QFrame.NoFrame)
        self.frame_7.setFrameShadow(QtWidgets.QFrame.Plain)
        self.frame_7.setObjectName("frame_7")
        self.horizontalLayout_8 = QtWidgets.QHBoxLayout(self.frame_7)
        self.horizontalLayout_8.setObjectName("horizontalLayout_8")
        spacerItem = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout_8.addItem(spacerItem)
        self.draw_by_peak_pushButton = QtWidgets.QPushButton(self.frame_7)
        self.draw_by_peak_pushButton.setText("")
        icon = QtGui.QIcon()
        icon.addPixmap(QtGui.QPixmap(":/Icons/icons/gear.svg"), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        self.draw_by_peak_pushButton.setIcon(icon)
        self.draw_by_peak_pushButton.setObjectName("draw_by_peak_pushButton")
        self.horizontalLayout_8.addWidget(self.draw_by_peak_pushButton)
        self.verticalLayout_4.addWidget(self.frame_7)
        spacerItem1 = QtWidgets.QSpacerItem(20, 40, QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Expanding)
        self.verticalLayout_4.addItem(spacerItem1)
        self.verticalLayout_3.addWidget(self.frame_2)
        self.toolBox.addItem(self.page_3, "")
        self.page = QtWidgets.QWidget()
        self.page.setGeometry(QtCore.QRect(0, 0, 830, 352))
        self.page.setObjectName("page")
        self.verticalLayout_2 = QtWidgets.QVBoxLayout(self.page)
        self.verticalLayout_2.setObjectName("verticalLayout_2")
        self.tableView = QtWidgets.QTableView(self.page)
        self.tableView.setObjectName("tableView")
        self.verticalLayout_2.addWidget(self.tableView)
        self.frame = QtWidgets.QFrame(self.page)
        self.frame.setMaximumSize(QtCore.QSize(16777215, 40))
        self.frame.setFrameShape(QtWidgets.QFrame.NoFrame)
        self.frame.setFrameShadow(QtWidgets.QFrame.Raised)
        self.frame.setObjectName("frame")
        self.horizontalLayout_2 = QtWidgets.QHBoxLayout(self.frame)
        self.horizontalLayout_2.setContentsMargins(0, 0, 0, 0)
        self.horizontalLayout_2.setObjectName("horizontalLayout_2")
        spacerItem2 = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout_2.addItem(spacerItem2)
        self.draw_by_points_pushButton = QtWidgets.QPushButton(self.frame)
        self.draw_by_points_pushButton.setText("")
        self.draw_by_points_pushButton.setIcon(icon)
        self.draw_by_points_pushButton.setObjectName("draw_by_points_pushButton")
        self.horizontalLayout_2.addWidget(self.draw_by_points_pushButton)
        self.verticalLayout_2.addWidget(self.frame)
        self.toolBox.addItem(self.page, "")
        self.verticalLayout_7.addWidget(self.toolBox)
        self.frame_9 = QtWidgets.QFrame(self.frame_8)
        self.frame_9.setMinimumSize(QtCore.QSize(0, 150))
        self.frame_9.setFrameShape(QtWidgets.QFrame.StyledPanel)
        self.frame_9.setFrameShadow(QtWidgets.QFrame.Raised)
        self.frame_9.setObjectName("frame_9")
        self.horizontalLayout_9 = QtWidgets.QHBoxLayout(self.frame_9)
        self.horizontalLayout_9.setObjectName("horizontalLayout_9")
        self.verticalLayout_7.addWidget(self.frame_9)
        self.PlotFrame = QtWidgets.QFrame(self.splitter_3)
        self.PlotFrame.setFrameShape(QtWidgets.QFrame.NoFrame)
        self.PlotFrame.setFrameShadow(QtWidgets.QFrame.Raised)
        self.PlotFrame.setObjectName("PlotFrame")
        self.horizontalLayout = QtWidgets.QHBoxLayout(self.PlotFrame)
        self.horizontalLayout.setObjectName("horizontalLayout")
        self.plotwidget = MatplotlibWidget(self.PlotFrame)
        self.plotwidget.setObjectName("plotwidget")
        self.horizontalLayout.addWidget(self.plotwidget)
        self.verticalLayout.addWidget(self.splitter_3)

        self.retranslateUi(Dialog)
        self.toolBox.setCurrentIndex(0)
        QtCore.QMetaObject.connectSlotsByName(Dialog)

    def retranslateUi(self, Dialog):
        Dialog.setWindowTitle(QtWidgets.QApplication.translate("Dialog", "Dialog", None, -1))
        self.label.setText(QtWidgets.QApplication.translate("Dialog", "Night valley (MW)", None, -1))
        self.label_2.setText(QtWidgets.QApplication.translate("Dialog", "Morning Peak (MW)", None, -1))
        self.label_4.setText(QtWidgets.QApplication.translate("Dialog", "Afternoon valley (MW)", None, -1))
        self.label_5.setText(QtWidgets.QApplication.translate("Dialog", "Evening Peak (MW)", None, -1))
        self.toolBox.setItemText(self.toolBox.indexOf(self.page_3), QtWidgets.QApplication.translate("Dialog", "Definition by peak points", None, -1))
        self.toolBox.setItemText(self.toolBox.indexOf(self.page), QtWidgets.QApplication.translate("Dialog", "Definition by data points", None, -1))

from .matplotlibwidget import MatplotlibWidget
from .icons_rc import *

if __name__ == "__main__":
    import sys
    app = QtWidgets.QApplication(sys.argv)
    Dialog = QtWidgets.QDialog()
    ui = Ui_Dialog()
    ui.setupUi(Dialog)
    Dialog.show()
    sys.exit(app.exec_())

